#include "f12.h"
#include "dco.hpp"
#include <random>

// This function is executed whenever the tape interpretor encounters
// an external adjoint object linked with it.
template <typename T, typename Tp, typename To>
void adjoint_bucket(typename dco::mode<T>::external_adjoint_object_t *EA) {
  // The tape due to external adjoint starts here. Any activity 
  // (e.g. interpretation) must not cross this boundary.
  typename dco::mode<T>::tape_t::position_t pos=dco::mode<T>::global_tape->get_position();
  // Local passive variables are recovered.
  int ib=EA->template read_data<int>(); int np=EA->template read_data<int>();
  int nb=EA->template read_data<int>(); int ns=EA->template read_data<int>();
  int nc=EA->template read_data<int>(); To dt=EA->template read_data<To>();
  // Local active variables are recovered.
  T x_c=EA->template read_data<T>(); T y=EA->template read_data<T>(); 
  // Pointer to read-only global active variables are recovered.
  std::vector<std::vector<Tp>>* p_p=EA->template read_data<std::vector<std::vector<Tp>>*>(); 
  // A tape of the current bucket of paths is generated. 
  // It is linked with the upstream tape through the active inputs
  // (here, x_c, y and parts of *p_p).
  for (int i=ib;i<ib+np/nb;++i) {
    T x=x_c;
    for (int jc=0;jc<ns;jc+=ns/nc) 
      for (int j=jc;j<jc+ns/nc;++j) {
        To Z=EA->template read_data<To>();
        x+=dt*f1(x,(*p_p)[0][j])+f2(x,(*p_p)[1][j])*sqrt(dt)*Z;
      }
    y+=x;
  }
  dco::derivative(y)=EA->get_output_adjoint();
  dco::mode<T>::global_tape->interpret_adjoint_and_reset_to(pos);
}

// The type-generic primal needs to be extended to 
// enable late linearization/accumulation of adjoints 
// of the target code (here, a single bucket of paths).
template <typename T, typename Tp, typename To=float>
T f(int np, int nb, int ns, int nc, T x, const std::vector<std::vector<Tp>>& p) {
  using namespace std;
  default_random_engine g; normal_distribution<To> d(0,1);
  T y=0, x_c=x; // read-only copy of x
  To dt=1./ns;
  for (int ib=0;ib<np;ib+=np/nb) {
    // External adjoint (EA) objects enable late 
    // linearization/accumulation of adjoints of 
    // the target code.
    typename dco::mode<T>::external_adjoint_object_t* EA=dco::mode<T>::global_tape->template create_callback_object<typename dco::mode<T>::external_adjoint_object_t>();
    // All inputs to the target code need to be accessible 
    // from adjoint_bucket; they are required variables. 
    // The local memory requirement should be minimal.
    //
    // Required local passive variables need to be stored.
    EA->write_data(ib); EA->write_data(np); EA->write_data(nb);
    EA->write_data(ns); EA->write_data(nc); EA->write_data(dt);
    // A local active variable needs to be stored if both its value and
    // its adjoint reference will be used in adjoint_bucket. Mere storage
    // of the value reduces the memory requirement in case the adjoint
    // reference is not required.
    EA->write_data(x_c); EA->write_data(y);
    // Read-only "global" variables (live at this point as well 
    // as when the corresponding call to adjoint_bucket is executed) 
    // can be access via their address in memory.
    EA->write_data(&p);
    // Computational efficiency of the passive target code may benefit from
    // copying into actual passive variables. Alternatively, taping 
    // can just be switched off.
    dco::mode<T>::global_tape->switch_to_passive();
    for (int i=ib;i<ib+np/nb;++i) {
      x=x_c;
      for (int jc=0;jc<ns;jc+=ns/nc) 
        for (int j=jc;j<jc+ns/nc;++j) {
          To Z=d(g); EA->write_data(Z);
          x+=dt*f1(x,p[0][j])+f2(x,p[1][j])*sqrt(dt)*Z;
        }
      y+=x;
    } 
    dco::mode<T>::global_tape->switch_to_active();
    // Active outputs from the bucket need to be registered with the EA.
    y=EA->register_output(dco::value(y));
    // The EA is linked with the adjoint_bucket callback function and it
    // is added to the tape.
    dco::mode<T>::global_tape->template insert_callback<typename dco::mode<T>::external_adjoint_object_t>(adjoint_bucket<T,Tp,To>,EA);
  }
  return pow(y/np,2);
}

int main(int argc, char* argv[]) {
  assert(argc==5); 
  int np=std::stoi(argv[1]), nb=std::stoi(argv[2]); 
  assert(np>0&&nb>0&&np%nb==0);
  int ns=std::stoi(argv[3]), nc=std::stoi(argv[4]); 
  assert(ns>0&&nc>0&&ns%nc==0);
  using Tb=double; std::cout.precision(std::numeric_limits<Tb>::digits10); 
  using Tp=dco::ga1s<Tb>::type; using T=Tp;
  T x=1.; std::vector<std::vector<Tp>> p(2,std::vector<Tp>(ns,1));
  dco::ga1s<Tb>::global_tape=dco::ga1s<Tb>::tape_t::create(); 
  dco::ga1s<Tb>::global_tape->register_variable(x); 
  for (int i=0;i<ns;i++) { 
    dco::ga1s<Tb>::global_tape->register_variable(p[0][i]); 
    dco::ga1s<Tb>::global_tape->register_variable(p[1][i]); 
  }
  dco::ga1s<Tb>::tape_t::position_t pos=dco::ga1s<Tb>::global_tape->get_position();
  T y=f(np,nb,ns,nc,x,p); 
  dco::ga1s<Tb>::global_tape->register_output_variable(y);
  dco::derivative(y)=1; 
  dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos); 
  std::cout << "dy/dx=" << dco::derivative(x) << '\n';
  for (int i=0;i<ns;i++) 
    std::cout << "dy/dp[0][" << i << "]=" << dco::derivative(p[0][i]) << '\n'
              << "dy/dp[1][" << i << "]=" << dco::derivative(p[1][i]) << std::endl;
  dco::ga1s<Tb>::tape_t::remove(dco::ga1s<Tb>::global_tape); 
  std::cout << "y=" << dco::value(y) << std::endl;
  return 0;
}

// second order working
/*

int main(int argc, char* argv[]) {
  assert(argc==5);
  int np=std::stoi(argv[1]), nb=std::stoi(argv[2]);
  assert(np>0&&nb>0&&np%nb==0);
  int ns=std::stoi(argv[3]), nc=std::stoi(argv[4]);
  assert(ns>0&&nc>0&&ns%nc==0);
  using Tbb=double; std::cout.precision(std::numeric_limits<Tbb>::digits10);
  using Tb=dco::gt1s<Tbb>::type; using Tp=dco::ga1s<Tb>::type; using T=Tp;
  T x=1,y; std::vector<std::vector<Tp>> p(2,std::vector<Tp>(ns,1));
  dco::ga1s<Tb>::global_tape=dco::ga1s<Tb>::tape_t::create(); 
  dco::ga1s<Tb>::global_tape->register_variable(x); 
  for (int i=0;i<ns;i++) { 
    dco::ga1s<Tb>::global_tape->register_variable(p[0][i]); 
    dco::ga1s<Tb>::global_tape->register_variable(p[1][i]); 
  }
  dco::ga1s<Tb>::tape_t::position_t pos=dco::ga1s<Tb>::global_tape->get_position();
  // d(dy/d(x,p))dx
  dco::derivative(dco::value(x))=1; 
  y=f(np,nb,ns,nc,x,p); 
  dco::value(dco::derivative(y))=1; 
  dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
  std::cout << "d(dy/dx)/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
  for (int j=0;j<ns;++j) {
    std::cout << "d(dy/dp[0][" << j << "])/dx=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
    dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
    std::cout << "d(dy/dp[1][" << j << "])/dx=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
    dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
  }
  dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
  // d(dy/d(x,p))dp
  for (int i=0;i<ns;++i) {
    dco::derivative(dco::value(p[0][i]))=1;
    y=f(np,nb,ns,nc,x,p); 
    dco::value(dco::derivative(y))=1; 
    dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
    std::cout << "d(dy/dp[0][" << i << "])/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
    dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
    for (int j=0;j<ns;++j) {
      std::cout << "d(dy/dp[0][" << i << "])/dp[0][" << j << "]=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
      dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
      std::cout << "d(dy/dp[0][" << i << "])/dp[1][" << j << "]=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
      dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
    }
    dco::derivative(dco::value(p[0][i]))=0;
    dco::derivative(dco::value(p[1][i]))=1;
    y=f(np,nb,ns,nc,x,p); 
    dco::value(dco::derivative(y))=1; 
    dco::ga1s<Tb>::global_tape->interpret_adjoint_and_reset_to(pos);
    std::cout << "d(dy/dp[1][" << i << "])/dx=" << dco::derivative(dco::derivative(x)) << std::endl;
    dco::derivative(dco::value(x))=0; dco::derivative(x)=0; 
    for (int j=0;j<ns;++j) {
      std::cout << "d(dy/dp[1][" << i << "])/dp[0][" << j << "]=" << dco::derivative(dco::derivative(p[0][j])) << std::endl;
      dco::derivative(dco::value(p[0][j]))=0; dco::derivative(p[0][j])=0;
      std::cout << "d(dy/dp[1][" << i << "])/dp[1][" << j << "]=" << dco::derivative(dco::derivative(p[1][j])) << std::endl;
      dco::derivative(dco::value(p[1][j]))=0; dco::derivative(p[1][j])=0;
    }
    dco::derivative(dco::value(p[1][i]))=0;
  }
  std::cout << "y=" << dco::value(dco::value(y)) << std::endl;
  return 0;
}

*/
